@@ -1,11 +1,12 @@
 // Copyright 2017, The Go Authors. All rights reserved.
 // Use of this source code is governed by a BSD-style
-// license that can be found in the LICENSE.md file.
+// license that can be found in the LICENSE file.
 
 package cmpopts
 
 import (
 	"bytes"
+	"errors"
 	"fmt"
 	"io"
 	"math"
@@ -16,11 +17,14 @@ import (
 	"time"
 
 	"github.com/google/go-cmp/cmp"
+	"golang.org/x/xerrors"
 )
 
 type (
 	MyInt    int
+	MyInts   []int
 	MyFloat  float32
+	MyString string
 	MyTime   struct{ time.Time }
 	MyStruct struct {
 		A, B []int
@@ -448,6 +452,231 @@ func TestOptions(t *testing.T) {
 		},
 		wantEqual: true,
 		reason:    "equal because named type is transformed to float64",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Date(2009, 11, 10, 23, 0, 0, 0, time.UTC),
+		y:         time.Date(2009, 11, 10, 23, 0, 0, 0, time.UTC),
+		opts:      []cmp.Option{EquateApproxTime(0)},
+		wantEqual: true,
+		reason:    "equal because times are identical",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Date(2009, 11, 10, 23, 0, 0, 0, time.UTC),
+		y:         time.Date(2009, 11, 10, 23, 0, 3, 0, time.UTC),
+		opts:      []cmp.Option{EquateApproxTime(3 * time.Second)},
+		wantEqual: true,
+		reason:    "equal because time is exactly at the allowed margin",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Date(2009, 11, 10, 23, 0, 3, 0, time.UTC),
+		y:         time.Date(2009, 11, 10, 23, 0, 0, 0, time.UTC),
+		opts:      []cmp.Option{EquateApproxTime(3 * time.Second)},
+		wantEqual: true,
+		reason:    "equal because time is exactly at the allowed margin (negative)",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Date(2009, 11, 10, 23, 0, 3, 0, time.UTC),
+		y:         time.Date(2009, 11, 10, 23, 0, 0, 0, time.UTC),
+		opts:      []cmp.Option{EquateApproxTime(3*time.Second - 1)},
+		wantEqual: false,
+		reason:    "not equal because time is outside allowed margin",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Date(2009, 11, 10, 23, 0, 0, 0, time.UTC),
+		y:         time.Date(2009, 11, 10, 23, 0, 3, 0, time.UTC),
+		opts:      []cmp.Option{EquateApproxTime(3*time.Second - 1)},
+		wantEqual: false,
+		reason:    "not equal because time is outside allowed margin (negative)",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Time{},
+		y:         time.Time{},
+		opts:      []cmp.Option{EquateApproxTime(3 * time.Second)},
+		wantEqual: true,
+		reason:    "equal because both times are zero",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Time{},
+		y:         time.Time{}.Add(1),
+		opts:      []cmp.Option{EquateApproxTime(3 * time.Second)},
+		wantEqual: false,
+		reason:    "not equal because zero time is always not equal not non-zero",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Time{}.Add(1),
+		y:         time.Time{},
+		opts:      []cmp.Option{EquateApproxTime(3 * time.Second)},
+		wantEqual: false,
+		reason:    "not equal because zero time is always not equal not non-zero",
+	}, {
+		label:     "EquateApproxTime",
+		x:         time.Date(2409, 11, 10, 23, 0, 0, 0, time.UTC),
+		y:         time.Date(2000, 11, 10, 23, 0, 3, 0, time.UTC),
+		opts:      []cmp.Option{EquateApproxTime(3 * time.Second)},
+		wantEqual: false,
+		reason:    "time difference overflows time.Duration",
+	}, {
+		label:     "EquateErrors",
+		x:         nil,
+		y:         nil,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "nil values are equal",
+	}, {
+		label:     "EquateErrors",
+		x:         errors.New("EOF"),
+		y:         io.EOF,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: false,
+		reason:    "user-defined EOF is not exactly equal",
+	}, {
+		label:     "EquateErrors",
+		x:         xerrors.Errorf("wrapped: %w", io.EOF),
+		y:         io.EOF,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "wrapped io.EOF is equal according to errors.Is",
+	}, {
+		label:     "EquateErrors",
+		x:         xerrors.Errorf("wrapped: %w", io.EOF),
+		y:         io.EOF,
+		wantEqual: false,
+		reason:    "wrapped io.EOF is not equal without EquateErrors option",
+	}, {
+		label:     "EquateErrors",
+		x:         io.EOF,
+		y:         io.EOF,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "sentinel errors are equal",
+	}, {
+		label:     "EquateErrors",
+		x:         io.EOF,
+		y:         AnyError,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "AnyError is equal to any non-nil error",
+	}, {
+		label:     "EquateErrors",
+		x:         io.EOF,
+		y:         AnyError,
+		wantEqual: false,
+		reason:    "AnyError is not equal to any non-nil error without EquateErrors option",
+	}, {
+		label:     "EquateErrors",
+		x:         nil,
+		y:         AnyError,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: false,
+		reason:    "AnyError is not equal to nil value",
+	}, {
+		label:     "EquateErrors",
+		x:         nil,
+		y:         nil,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "nil values are equal",
+	}, {
+		label:     "EquateErrors",
+		x:         errors.New("EOF"),
+		y:         io.EOF,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: false,
+		reason:    "user-defined EOF is not exactly equal",
+	}, {
+		label:     "EquateErrors",
+		x:         xerrors.Errorf("wrapped: %w", io.EOF),
+		y:         io.EOF,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "wrapped io.EOF is equal according to errors.Is",
+	}, {
+		label:     "EquateErrors",
+		x:         xerrors.Errorf("wrapped: %w", io.EOF),
+		y:         io.EOF,
+		wantEqual: false,
+		reason:    "wrapped io.EOF is not equal without EquateErrors option",
+	}, {
+		label:     "EquateErrors",
+		x:         io.EOF,
+		y:         io.EOF,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "sentinel errors are equal",
+	}, {
+		label:     "EquateErrors",
+		x:         io.EOF,
+		y:         AnyError,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "AnyError is equal to any non-nil error",
+	}, {
+		label:     "EquateErrors",
+		x:         io.EOF,
+		y:         AnyError,
+		wantEqual: false,
+		reason:    "AnyError is not equal to any non-nil error without EquateErrors option",
+	}, {
+		label:     "EquateErrors",
+		x:         nil,
+		y:         AnyError,
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: false,
+		reason:    "AnyError is not equal to nil value",
+	}, {
+		label:     "EquateErrors",
+		x:         struct{ E error }{nil},
+		y:         struct{ E error }{nil},
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "nil values are equal",
+	}, {
+		label:     "EquateErrors",
+		x:         struct{ E error }{errors.New("EOF")},
+		y:         struct{ E error }{io.EOF},
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: false,
+		reason:    "user-defined EOF is not exactly equal",
+	}, {
+		label:     "EquateErrors",
+		x:         struct{ E error }{xerrors.Errorf("wrapped: %w", io.EOF)},
+		y:         struct{ E error }{io.EOF},
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "wrapped io.EOF is equal according to errors.Is",
+	}, {
+		label:     "EquateErrors",
+		x:         struct{ E error }{xerrors.Errorf("wrapped: %w", io.EOF)},
+		y:         struct{ E error }{io.EOF},
+		wantEqual: false,
+		reason:    "wrapped io.EOF is not equal without EquateErrors option",
+	}, {
+		label:     "EquateErrors",
+		x:         struct{ E error }{io.EOF},
+		y:         struct{ E error }{io.EOF},
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "sentinel errors are equal",
+	}, {
+		label:     "EquateErrors",
+		x:         struct{ E error }{io.EOF},
+		y:         struct{ E error }{AnyError},
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: true,
+		reason:    "AnyError is equal to any non-nil error",
+	}, {
+		label:     "EquateErrors",
+		x:         struct{ E error }{io.EOF},
+		y:         struct{ E error }{AnyError},
+		wantEqual: false,
+		reason:    "AnyError is not equal to any non-nil error without EquateErrors option",
+	}, {
+		label:     "EquateErrors",
+		x:         struct{ E error }{nil},
+		y:         struct{ E error }{AnyError},
+		opts:      []cmp.Option{EquateErrors()},
+		wantEqual: false,
+		reason:    "AnyError is not equal to nil value",
 	}, {
 		label:     "IgnoreFields",
 		x:         Bar1{Foo3{&Foo2{&Foo1{Alpha: 5}}}},
@@ -542,6 +771,39 @@ func TestOptions(t *testing.T) {
 		opts:      []cmp.Option{IgnoreFields(Bar3{}, "Bar1", "Bravo", "Delta", "Alpha")},
 		wantEqual: false,
 		reason:    "not equal because highest-level field is not ignored: Foo3",
+	}, {
+		label: "IgnoreFields",
+		x: ParentStruct{
+			privateStruct: &privateStruct{private: 1},
+			PublicStruct:  &PublicStruct{private: 2},
+			private:       3,
+		},
+		y: ParentStruct{
+			privateStruct: &privateStruct{private: 10},
+			PublicStruct:  &PublicStruct{private: 20},
+			private:       30,
+		},
+		opts:      []cmp.Option{cmp.AllowUnexported(ParentStruct{}, PublicStruct{}, privateStruct{})},
+		wantEqual: false,
+		reason:    "not equal because unexported fields mismatch",
+	}, {
+		label: "IgnoreFields",
+		x: ParentStruct{
+			privateStruct: &privateStruct{private: 1},
+			PublicStruct:  &PublicStruct{private: 2},
+			private:       3,
+		},
+		y: ParentStruct{
+			privateStruct: &privateStruct{private: 10},
+			PublicStruct:  &PublicStruct{private: 20},
+			private:       30,
+		},
+		opts: []cmp.Option{
+			cmp.AllowUnexported(ParentStruct{}, PublicStruct{}, privateStruct{}),
+			IgnoreFields(ParentStruct{}, "PublicStruct.private", "privateStruct.private", "private"),
+		},
+		wantEqual: true,
+		reason:    "equal because mismatching unexported fields are ignored",
 	}, {
 		label:     "IgnoreTypes",
 		x:         []interface{}{5, "same"},
@@ -716,10 +978,111 @@ func TestOptions(t *testing.T) {
 		},
 		wantEqual: true,
 		reason:    "equal because all Ignore options can be composed together",
+	}, {
+		label: "IgnoreSliceElements",
+		x:     []int{1, 0, 2, 3, 0, 4, 0, 0},
+		y:     []int{0, 0, 0, 0, 1, 2, 3, 4},
+		opts: []cmp.Option{
+			IgnoreSliceElements(func(v int) bool { return v == 0 }),
+		},
+		wantEqual: true,
+		reason:    "equal because zero elements are ignored",
+	}, {
+		label: "IgnoreSliceElements",
+		x:     []MyInt{1, 0, 2, 3, 0, 4, 0, 0},
+		y:     []MyInt{0, 0, 0, 0, 1, 2, 3, 4},
+		opts: []cmp.Option{
+			IgnoreSliceElements(func(v int) bool { return v == 0 }),
+		},
+		wantEqual: false,
+		reason:    "not equal because MyInt is not assignable to int",
+	}, {
+		label: "IgnoreSliceElements",
+		x:     MyInts{1, 0, 2, 3, 0, 4, 0, 0},
+		y:     MyInts{0, 0, 0, 0, 1, 2, 3, 4},
+		opts: []cmp.Option{
+			IgnoreSliceElements(func(v int) bool { return v == 0 }),
+		},
+		wantEqual: true,
+		reason:    "equal because the element type of MyInts is assignable to int",
+	}, {
+		label: "IgnoreSliceElements+EquateEmpty",
+		x:     []MyInt{},
+		y:     []MyInt{0, 0, 0, 0},
+		opts: []cmp.Option{
+			IgnoreSliceElements(func(v int) bool { return v == 0 }),
+			EquateEmpty(),
+		},
+		wantEqual: false,
+		reason:    "not equal because ignored elements does not imply empty slice",
+	}, {
+		label: "IgnoreMapEntries",
+		x:     map[string]int{"one": 1, "TWO": 2, "three": 3, "FIVE": 5},
+		y:     map[string]int{"one": 1, "three": 3, "TEN": 10},
+		opts: []cmp.Option{
+			IgnoreMapEntries(func(k string, v int) bool { return strings.ToUpper(k) == k }),
+		},
+		wantEqual: true,
+		reason:    "equal because uppercase keys are ignored",
+	}, {
+		label: "IgnoreMapEntries",
+		x:     map[MyString]int{"one": 1, "TWO": 2, "three": 3, "FIVE": 5},
+		y:     map[MyString]int{"one": 1, "three": 3, "TEN": 10},
+		opts: []cmp.Option{
+			IgnoreMapEntries(func(k string, v int) bool { return strings.ToUpper(k) == k }),
+		},
+		wantEqual: false,
+		reason:    "not equal because MyString is not assignable to string",
+	}, {
+		label: "IgnoreMapEntries",
+		x:     map[string]MyInt{"one": 1, "TWO": 2, "three": 3, "FIVE": 5},
+		y:     map[string]MyInt{"one": 1, "three": 3, "TEN": 10},
+		opts: []cmp.Option{
+			IgnoreMapEntries(func(k string, v int) bool { return strings.ToUpper(k) == k }),
+		},
+		wantEqual: false,
+		reason:    "not equal because MyInt is not assignable to int",
+	}, {
+		label: "IgnoreMapEntries+EquateEmpty",
+		x:     map[string]MyInt{"ONE": 1, "TWO": 2, "THREE": 3},
+		y:     nil,
+		opts: []cmp.Option{
+			IgnoreMapEntries(func(k string, v int) bool { return strings.ToUpper(k) == k }),
+			EquateEmpty(),
+		},
+		wantEqual: false,
+		reason:    "not equal because ignored entries does not imply empty map",
+	}, {
+		label: "AcyclicTransformer",
+		x:     "a\nb\nc\nd",
+		y:     "a\nb\nd\nd",
+		opts: []cmp.Option{
+			AcyclicTransformer("", func(s string) []string { return strings.Split(s, "\n") }),
+		},
+		wantEqual: false,
+		reason:    "not equal because 3rd line differs, but should not recurse infinitely",
+	}, {
+		label: "AcyclicTransformer",
+		x:     []string{"foo", "Bar", "BAZ"},
+		y:     []string{"Foo", "BAR", "baz"},
+		opts: []cmp.Option{
+			AcyclicTransformer("", strings.ToUpper),
+		},
+		wantEqual: true,
+		reason:    "equal because of strings.ToUpper; AcyclicTransformer unnecessary, but check this still works",
+	}, {
+		label: "AcyclicTransformer",
+		x:     "this is a sentence",
+		y: "this   			is a 			sentence",
+		opts: []cmp.Option{
+			AcyclicTransformer("", strings.Fields),
+		},
+		wantEqual: true,
+		reason:    "equal because acyclic transformer splits on any contiguous whitespace",
 	}}
 
 	for _, tt := range tests {
-		tRun(t, tt.label, func(t *testing.T) {
+		t.Run(tt.label, func(t *testing.T) {
 			var gotEqual bool
 			var gotPanic string
 			func() {
@@ -731,6 +1094,8 @@ func TestOptions(t *testing.T) {
 				gotEqual = cmp.Equal(tt.x, tt.y, tt.opts...)
 			}()
 			switch {
+			case tt.reason == "":
+				t.Errorf("reason must be provided")
 			case gotPanic == "" && tt.wantPanic:
 				t.Errorf("expected Equal panic\nreason: %s", tt.reason)
 			case gotPanic != "" && !tt.wantPanic:
@@ -783,6 +1148,12 @@ func TestPanic(t *testing.T) {
 		fnc:    EquateApprox,
 		args:   args(0.0, math.Inf(+1)),
 		reason: "margin of infinity is valid",
+	}, {
+		label:     "EquateApproxTime",
+		fnc:       EquateApproxTime,
+		args:      args(time.Duration(-1)),
+		wantPanic: "margin must be a non-negative number",
+		reason:    "negative duration is invalid",
 	}, {
 		label:     "SortSlices",
 		fnc:       SortSlices,
@@ -852,14 +1223,24 @@ func TestPanic(t *testing.T) {
 		label:     "IgnoreFields",
 		fnc:       IgnoreFields,
 		args:      args(&Foo1{}, "Alpha"),
-		wantPanic: "must be a struct",
+		wantPanic: "must be a non-pointer struct",
 		reason:    "the type must be a struct (not pointer to a struct)",
+	}, {
+		label:  "IgnoreFields",
+		fnc:    IgnoreFields,
+		args:   args(struct{ privateStruct }{}, "privateStruct"),
+		reason: "privateStruct field permitted since it is the default name of the embedded type",
+	}, {
+		label:  "IgnoreFields",
+		fnc:    IgnoreFields,
+		args:   args(struct{ privateStruct }{}, "Public"),
+		reason: "Public field permitted since it is a forwarded field that is exported",
 	}, {
 		label:     "IgnoreFields",
 		fnc:       IgnoreFields,
-		args:      args(Foo1{}, "unexported"),
-		wantPanic: "name must be exported",
-		reason:    "unexported fields must not be specified",
+		args:      args(struct{ privateStruct }{}, "private"),
+		wantPanic: "does not exist",
+		reason:    "private field not permitted since it is a forwarded field that is unexported",
 	}, {
 		label:  "IgnoreTypes",
 		fnc:    IgnoreTypes,
@@ -923,23 +1304,29 @@ func TestPanic(t *testing.T) {
 		label:     "IgnoreUnexported",
 		fnc:       IgnoreUnexported,
 		args:      args(nil),
-		wantPanic: "invalid struct type",
+		wantPanic: "must be a non-pointer struct",
 		reason:    "input must not be nil value",
 	}, {
 		label:     "IgnoreUnexported",
 		fnc:       IgnoreUnexported,
 		args:      args(&Foo1{}),
-		wantPanic: "invalid struct type",
+		wantPanic: "must be a non-pointer struct",
 		reason:    "input must be a struct type (not a pointer to a struct)",
 	}, {
 		label:  "IgnoreUnexported",
 		fnc:    IgnoreUnexported,
 		args:   args(Foo1{}, struct{ x, X int }{}),
 		reason: "input may be named or unnamed structs",
+	}, {
+		label:     "AcyclicTransformer",
+		fnc:       AcyclicTransformer,
+		args:      args("", "not a func"),
+		wantPanic: "invalid transformer function",
+		reason:    "AcyclicTransformer has same input requirements as Transformer",
 	}}
 
 	for _, tt := range tests {
-		tRun(t, tt.label, func(t *testing.T) {
+		t.Run(tt.label, func(t *testing.T) {
 			// Prepare function arguments.
 			vf := reflect.ValueOf(tt.fnc)
 			var vargs []reflect.Value
@@ -972,6 +1359,8 @@ func TestPanic(t *testing.T) {
 			}()
 
 			switch {
+			case tt.reason == "":
+				t.Errorf("reason must be provided")
 			case tt.wantPanic == "" && gotPanic != "":
 				t.Errorf("unexpected panic message: %s\nreason: %s", gotPanic, tt.reason)
 			case tt.wantPanic != "" && !strings.Contains(gotPanic, tt.wantPanic):
@@ -980,17 +1369,3 @@ func TestPanic(t *testing.T) {
 		})
 	}
 }
-
-// TODO: Delete this hack when we drop Go1.6 support.
-func tRun(t *testing.T, name string, f func(t *testing.T)) {
-	type runner interface {
-		Run(string, func(t *testing.T)) bool
-	}
-	var ti interface{} = t
-	if r, ok := ti.(runner); ok {
-		r.Run(name, f)
-	} else {
-		t.Logf("Test: %s", name)
-		f(t)
-	}
-}
